unit vision;

interface

uses
  matrices, sysutils, rapport, math;

type
  fittingfn = function(pts1, pts2, H: TMatrix; Tol: Real): Integer;
  distfn = procedure(H, X: TMatrix; d: TVector);
  degenfn = function(X: TMatrix): Integer;

procedure normalise2dpts(pts, newpts, T: TMatrix);
function homography2d(pts1, pts2, H: TMatrix; Tol: Real): Integer;
function homography2dnlin(pts1, pts2, H0, H: TMatrix; maxiter: Integer; var fi, ff: Real): Integer;
function ransacfithomography(pts1, pts2, ptsin1, ptsin2, H: TMatrix; tolransac: Real; var ninliers: Integer): Integer;
procedure homogdist2d(H, X: TMatrix; d: TVector);
function ransac(X: TMatrix; fit: fittingfn; dist: distfn; deg: degenfn; s: Integer; t: Real; H: TMatrix; inliers: TVector; var ninliers: Integer): Integer;

const

  RAN_OK = 0;
  RAN_DEG = -1;
  RAN_MAX = -2;
  RAN_NO = -3;

implementation

uses
  fmath, optim;

var
  locpts1, locpts2: TMatrix;

function iscolinear(x1,y1,x2,y2,x3,y3: Real): Integer;
var
  dx1, dx2, dy1, dy2: Real;
begin
  dx1:= x2 - x1;
  dx2:= x3 - x1;
  dy1:= y2 - y1;
  dy2:= y3 - y1;
  if ((dx1*dy2 - dy1*dx2) = 0) then Result:= 1
  else Result:= 0;
end;

function isdegenerate(X: TMatrix): Integer;
begin
  Result:= iscolinear(X[0][0],X[1][0],X[0][1],X[1][1],X[0][2],X[1][2]) +
           iscolinear(X[0][0],X[1][0],X[0][2],X[1][2],X[0][3],X[1][3]) +
           iscolinear(X[0][1],X[1][1],X[0][2],X[1][2],X[0][3],X[1][3]) +
           iscolinear(X[0][0],X[1][0],X[0][1],X[1][1],X[0][3],X[1][3]) +

           iscolinear(X[3][0],X[4][0],X[3][1],X[4][1],X[3][2],X[4][2]) +
           iscolinear(X[3][0],X[4][0],X[3][2],X[4][2],X[3][3],X[4][3]) +
           iscolinear(X[3][1],X[4][1],X[3][2],X[4][2],X[3][3],X[4][3]) +
           iscolinear(X[3][0],X[4][0],X[3][1],X[4][1],X[3][3],X[4][3]);
end;

procedure hnormalise(V1, V2: TMatrix);
var
  i: Integer;
begin
  for i:= Low(V1[0]) to High(V1[0]) do begin
    V2[0][i]:= V1[0][i] / V1[2][i];
    V2[1][i]:= V1[1][i] / V1[2][i];
    V2[2][i]:= 1;
  end;
end;

procedure homogdist2d(H, X: TMatrix; d: TVector);
var
  x1, x2, Hx1, invHx2, Ht: TMatrix;
  i, n: Integer;
begin
  n:= Length(X[0]);
  DimMatrix(x1, 2, n-1);
  DimMatrix(x2, 2, n-1);
  DimMatrix(Hx1, 2, n-1);
  DimMatrix(invHx2, 2, n-1);
  DimMatrix(Ht, 2, 2);

  for i:= 0 to 2 do begin
    CopyVector(x1[i], X[i], 0, n-1);
    CopyVector(x2[i], X[i+3], 0, n-1);
  end;
  MatMult(H, x1, Hx1);
  InvMat(H, 0, 2, Ht);
  MatMult(Ht, x2, invHx2);

  hnormalise(x1, x1);
  hnormalise(x2, x2);
  hnormalise(Hx1, Hx1);
  hnormalise(invHx2, invHx2);

  for i:= 0 to (n-1) do begin
    d[i]:= power((x1[0][i]-invHx2[0][i]),2) + power((x1[1][i]-invHx2[1][i]),2) + power((x1[2][i]-invHx2[2][i]),2)
         + power((x2[0][i]-Hx1[0][i]),2) + power((x2[1][i]-Hx1[1][i]),2) + power((x2[2][i]-Hx1[2][i]),2);
  end;
end;

function FObj(X : TVector) : Float;
var
  x147, x258, x369, yu, yv: Real;
  i: Integer;
begin
  Result:= 0;
  for i:= 0 to Length(locpts1[0])-1 do begin
    x147:= X[0]*locpts1[0][i] + X[3]*locpts1[1][i] + X[6];
    x258:= X[1]*locpts1[0][i] + X[4]*locpts1[1][i] + X[7];
    x369:= X[2]*locpts1[0][i] + X[5]*locpts1[1][i] + X[8];
    yu:= locpts2[0][i] - x147/x369;
    yv:= locpts2[1][i] - x258/x369;
    Result:= Result + sqrt(yu*yu+yv*yv);
  end;
end;

function homography2dnlin(pts1, pts2, H0, H: TMatrix; maxiter: Integer; var fi, ff: Real): Integer;
var
  res: Integer;
  F_min: Float;
  H_inv, Ht: TMatrix;
  XV: TVector;
begin
  locpts1:= pts1;
  locpts2:= pts2;
  DimVector(XV, 8);
  DimMatrix(H_inv, 8, 8);
  DimMatrix(Ht, 2, 2);

  Transpose(H0, 0, 0, 2, 2, Ht);
  MatToVect(Ht, XV);
  fi:= FObj(XV);
  res:= Marquardt(FObj, NumHessGrad, XV, 0, 8, maxiter, 1E-10, F_min, H_inv);
  ff:= F_min;
  VectToMat(XV, Ht);
  Transpose(Ht, 0, 0, 2, 2, H);
end;

function homography2d(pts1, pts2, H: TMatrix; Tol: Real): Integer;
var
  x1, x2, T1, T2, A, Tp: TMatrix;
  npts, l, p: Integer;
  x, y, w: Real;
  S: TVector;
  V: TMatrix;
begin
  npts:= Length(pts1[0]);
  DimMatrix(x1, 2, npts-1);
  DimMatrix(x2, 2, npts-1);
  DimMatrix(T1, 2, 2);
  DimMatrix(T2, 2, 2);
  DimMatrix(Tp, 2, 2);
  normalise2dpts(pts1, x1, T1);
  normalise2dpts(pts2, x2, T2);

  DimMatrix(A, 3*npts-1, 8);
  for l:= 0 to npts-1 do begin
    x:= x2[0][l];
    y:= x2[1][l];
    w:= x2[2][l];
    A[3*l][0]:= 0;
    A[3*l][1]:= 0;
    A[3*l][2]:= 0;
    A[3*l][3]:= -w*x1[0][l];
    A[3*l][4]:= -w*x1[1][l];
    A[3*l][5]:= -w*x1[2][l];
    A[3*l][6]:= y*x1[0][l];
    A[3*l][7]:= y*x1[1][l];
    A[3*l][8]:= y*x1[2][l];

    A[3*l+1][0]:= w*x1[0][l];
    A[3*l+1][1]:= w*x1[1][l];
    A[3*l+1][2]:= w*x1[2][l];
    A[3*l+1][3]:= 0;
    A[3*l+1][4]:= 0;
    A[3*l+1][5]:= 0;
    A[3*l+1][6]:= -x*x1[0][l];
    A[3*l+1][7]:= -x*x1[1][l];
    A[3*l+1][8]:= -x*x1[2][l];

    A[3*l+2][0]:= -y*x1[0][l];
    A[3*l+2][1]:= -y*x1[1][l];
    A[3*l+2][2]:= -y*x1[2][l];
    A[3*l+2][3]:= x*x1[0][l];
    A[3*l+2][4]:= x*x1[1][l];
    A[3*l+2][5]:= x*x1[2][l];
    A[3*l+2][6]:= 0;
    A[3*l+2][7]:= 0;
    A[3*l+2][8]:= 0;
  end;

  DimVector(S, 8);
  DimMatrix(V, 8, 8);

  Result:= SV_Decomp(A, 0, 3*npts-1, 8, S, V);
  if (Result <> MAT_OK) then Exit;

  SV_SetZero(S, 0, 8, Tol);

  { prendre le vecteur de V correspondant  la plus petite valeur propre
      dans S }
  w:= S[0];
  p:= 0;
  for l:= 0 to 8 do begin
    if (S[l] <= w) then begin
      w:= S[l];
      p:= l;
    end;
  end;
  for l:= 0 to 8 do
    H[Floor(l/3)][l mod 3]:= V[l][p];
  InvMat(T2, 0, 2, Tp);
  MatMult(Tp, H, T2);
  MatMult(T2, T1, H);
end;

procedure normalise2dpts(pts, newpts, T: TMatrix);
var
  n: Integer;
  c1, c2, scale: Real;
  Vt: TVector;
begin
  DimVector(Vt, High(pts[0]));

  VectPDiv(pts[0], pts[2], pts[0]);
  VectPDiv(pts[1], pts[2], pts[1]);

  c1:= VectMean(pts[0]);
  c2:= VectMean(pts[1]);

  VectDiff(pts[0], newpts[0], c1);
  VectDiff(pts[1], newpts[1], c2);

  VectPMult(newpts[0], newpts[0], newpts[0]);
  VectPMult(newpts[1], newpts[1], newpts[1]);
  VectPAdd(newpts[0], newpts[1], Vt);
  VectSqrt(Vt, Vt);
  scale:= sqrt(2) / VectMean(Vt);

  MatFill(T, [scale,0,-scale*c1,0,scale,-scale*c2,0,0,1]);
  MatMult(T, pts, newpts);
end;

function ransac(X: TMatrix; fit: fittingfn; dist: distfn; deg: degenfn; s: Integer; t: Real; H: TMatrix; inliers: TVector; var ninliers: Integer): Integer;
var
  rows, npts, i, ind: Integer;
  p: Real;
  maxTrials, maxDataTrials: Integer;
  trialcount, degenerate, count: Integer;
  bestscore: Integer;
  Xt, Sa, Sat, pts1, pts2: TMatrix;
  d: TVector;
  bestH: TMatrix;
  fracinliers, pNoOutliers, N: Real;
  found: Boolean;
  bestInliers: TVector;
begin
  Result:= RAN_OK;
  rows:= Length(X);
  npts:= Length(X[0]);
  p:= 0.99;
  maxTrials:= 1000;
  maxDataTrials:= 100;
  trialcount:= 0;
  bestscore:=  0;
  N:= 1;

  found:= false;
  DimMatrix(bestH, 2, 2);
  DimMatrix(Xt, npts-1, 5);
  DimMatrix(Sat, s-1, 5);
  DimMatrix(Sa, 5, s-1);
  DimMatrix(pts1, 2, s-1);
  DimMatrix(pts2, 2, s-1);
  DimVector(d, npts-1);
  DimVector(bestinliers, npts-1);
  Transpose(X, 0, 0, 5, npts-1, Xt);
  while (N > trialcount) do begin
    degenerate:= 1;
    count:= 1;
    while (degenerate > 0) do begin
      for i:= 0 to s-1 do begin
        ind:= Random(npts);
        Sat[i]:= Xt[ind];
      end;
      Transpose(Sat, 0, 0, s-1, 5, Sa);
      degenerate:= deg(Sa);

      count:= count + 1;
      if (count > maxDataTrials) then begin
        //warning('Unable to select a nondegenerate data set');
        Result:= RAN_DEG;
        Exit;
      end;
    end;

    for i:= 0 to 2 do begin
      pts1[i]:= Sa[i];
      pts2[i]:= Sa[i+3];
    end;

    fit(pts1, pts2, H, 1.0E-8);
    dist(H, X, d);

    ninliers:= 0;
    for i:= 0 to npts-1 do begin
      if (abs(d[i]) < t) then begin
        inliers[ninliers]:= i;
        Inc(ninliers);
      end;
    end;

    if (ninliers > bestscore) then begin
      found:= true;
      bestscore:= ninliers;
      CopyVector(bestInliers, inliers, 0, ninliers-1);
      CopyMatrix(bestH,H,0,0,2,2);
      fracinliers:= ninliers/npts;
      pNoOutliers:= 1 -  power(fracinliers,s);
      pNoOutliers:= max(eps, pNoOutliers);
      pNoOutliers:= min(1-eps, pNoOutliers);
      N:= log(1-p)/log(pNoOutliers);
    end;

    trialcount:= trialcount+1;
    //fprintf('trial %d out of %d         \r',trialcount, ceil(N));

    if (trialcount > maxTrials) then begin
        //warning( ...
        //sprintf('ransac reached the maximum number of %d trials',...
        //maxTrials));
      Result:= RAN_MAX;
      Break;
    end;
  end;

  if (found) then begin
    Result:= 0;
    CopyMatrix(H, BestH, 0, 0, 2, 2);
    ninliers:= bestscore;
    CopyVector(inliers, bestInliers, 0, ninliers-1);
    //Transpose(bestinliers, 0, 0, npts-1, 5, inliers);
  end
  else
    Result:= RAN_NO;
end;

function ransacfithomography(pts1, pts2, ptsin1, ptsin2, H: TMatrix; tolransac: Real; var ninliers: Integer): Integer;
var
  rows, npts, i: Integer;
  x1, x2, T1, T2, X, Tp: TMatrix;
  s: Integer;
  Vt: TVector;
  inliers: TVector;
  xin1, xin2: TMatrix;
begin
  rows:= Length(pts1);
  npts:= Length(pts1[0]);

  if (npts < 4) then begin
    Result:= MAT_BAD_PARAM;
    Exit;
  end;

  DimMatrix(x1, 2, npts-1);
  DimMatrix(x2, 2, npts-1);
  DimMatrix(T1, 2, 2);
  DimMatrix(T2, 2, 2);
  normalise2dpts(pts1, x1, T1);
  normalise2dpts(pts2, x2, T2);


  s:= 4;
  DimMatrix(X, 5, Length(pts1)-1);
  X[0]:= x1[0];
  X[1]:= x1[1];
  X[2]:= x1[2];
  X[3]:= x2[0];
  X[4]:= x2[1];
  X[5]:= X2[2];
  DimVector(inliers, npts-1);
  ransac(X, homography2d, homogdist2d, isdegenerate, s, tolransac, H, inliers, ninliers);

  DimMatrix(xin1, 2, ninliers-1);
  DimMatrix(xin2, 2, ninliers-1);
  DimVector(Vt, 2);
  for i:= 0 to ninliers-1 do begin
    CopyVectorFromCol(Vt, x1, 0, 2, Floor(inliers[i]));
    CopyColFromVector(xin1, Vt, 0, 2, i);
    CopyVectorFromCol(Vt, x2, 0, 2, Floor(inliers[i]));
    CopyColFromVector(xin2, Vt, 0, 2, i);
  end;

  homography2d(xin1, xin2, H, 1e-8);

  DimMatrix(Tp, 2, 2);
  InvMat(T2, 0, 2, Tp);
  MatMult(Tp, H, T2);
  MatMult(T2, T1, H);

  for i:= 0 to ninliers-1 do begin
    CopyVectorFromCol(Vt, pts1, 0, 2, Floor(inliers[i]));
    CopyColFromVector(ptsin1, Vt, 0, 2, i);
    CopyVectorFromCol(Vt, pts2, 0, 2, Floor(inliers[i]));
    CopyColFromVector(ptsin2, Vt, 0, 2, i);
  end;
end;

end.
